var debug;
var delim = "__delim__";


// Nodedefs (.def)

var NodedefsView = Backbone.View.extend({
    el: $("#nodelistContainer"),
    initialize: function () {
        this.listenTo(Nodedefs, 'add', this.addNodedef);
        Nodedefs.trigger("fetch", "/nodedefs?def=data-flow");
        Nodedefs.trigger("fetch", "/nodedefs?def=hark");
        $("#nodelistFlat").css("display", "none");
        $("#nodeNotFound").css("display", "none");
        App.trigger("resized");
    },
    addNodedef: function (node) {
        this.addNodedef_sub(node, "#nodelist", false);
        this.addNodedef_sub(node, "#nodelistFlat", true);
    },

    addNodedef_sub: function (node, id, isflat) {
        var view = (new NodedefView({model: node})).render();
        var categories = this.$el.children(id).children("li").children("a");
        var target = _.find(categories, function (a) {
            return $(a).text().trim() === node.get("category");
        });

        if (!target) {
            this.$el.children(id).append(
                _.template(
                "<li name='<%= category %>''><a href='#'><%= category %></a><ul></ul></li>",
                node.toJSON()));
            categories = this.$el.children(id).children("li").children("a");
            target = _.find(categories, function (a) {
                return $(a).text().trim() === node.get("category");
            });
        }
        $(target).siblings("ul").append(view.el);
        if (isflat) {
            var view2 = (new NodedefView({model: node})).render();
            this.$el.children(id).append(view2.el);
        }
        if (this.$el.children(id).hasClass("ui-menu")) {
            this.$el.children(id).menu("refresh");
        }
        else {
            this.$el.children(id).menu();
        }
    }
});
var NodedefSearchView = Backbone.View.extend({
    el: $("#nodelistSearch"),
    events:{
        "keyup input": "search"
    },
    search: function () {
        var query = this.$el.children("input").val();

        if (query === "") {
            $("#nodelistFlat").css("display", "none");
            $("#nodelist").css("display", "block");
        }else {
            $("#nodelist").css("display", "none");
            $("#nodelistFlat").css("display", "block");
        }

        var isNotFound = true;
        _.each($("#nodelistFlat > li"), function (li) {
            var tmp = $(li).attr("name").toLowerCase();
            if (tmp.indexOf(query) == -1) {
                $(li).fadeOut("fast");
            }else {
            isNotFound = false;
                $(li).fadeIn("fast");
            }
        });
        if (isNotFound)
            $("#nodeNotFound").css("display", "block");
        else
            $("#nodeNotFound").css("display", "none");
    }
})

var NodedefView = Backbone.View.extend({
    tagName: "li",
    attributes: function () {
        return {"name": this.model.get("name")};
    },
    render: function() {
        this.$el.html("<a href='#'>" + this.model.get("name") + "</a>");
        return this;
    },
    events:{
        "click": "addNode",
    },
    addNode: function() {
        var nodedef = Nodedefs.find(function(def) {
            return def.get("name") === this.model.get("name");
        }, this);
        Sheets.trigger("addNodeToActiveSheet", nodedef);
    }
});


// Sheets

var SheetsView = Backbone.View.extend({
    el: $("#sheets"),
    initialize: function() {
        this.listenTo(Sheets, 'add', this.addSheet);
        this.listenTo(App, "resized", this.resize);
        Sheets.add({name: "MAIN", type: "subnet"});
        Sheets.trigger("activate", "MAIN");
        this.$el.contextMenu({
            selector: ".node",
            items: {
                "foo": {
                    name: "Foo",
                    icon: "edit",
                    callback: function(key, opt){
                        console.log("key", key);
                        console.log("opt", opt);
                    }
                },
                "bar": {
                    name: "Bar",
                    callback: function(key, opt){
                        console.log("key", key);
                        console.log("opt", opt);
                    }
                }
            }
        });
    },
    addSheet: function(sheet) {
        var nameView = (new SheetNameView({model: sheet})).render();
        var bodyView = (new SheetBodyView({model: sheet})).render();
        var nodesView= (new NodesView({collection: sheet.get("nodes")})).render();
        var linksView= (new LinksView({collection: sheet.get("links")})).render();
        var rolesView= (new RolesView({collection: sheet.get("roles")})).render();

        sheet.listenTo(sheet.get("nodes"), "change",
                       function(){linksView.updatePos(sheet.get("nodes"));});

        $(this.$el.children("ul").children("li")).eq(-1).before(nameView.el);
        $(this.$el).append(bodyView.el);
        $("#" + sheet.get("name") + "_sheet").append(nodesView.el);
        $("#" + sheet.get("name") + "_sheet").append(linksView.el);


        if (this.$el.hasClass("ui-tabs")) {
            this.$el.tabs("refresh");
        }
        else {
            this.$el.tabs({
                beforeActivate: function(e, ui) {
                    Sheets.trigger("activate", $(ui.newTab).text().trim());
            }
          });
        }
    },

    resize: function() {
        var that = this;
        // resize function
        this.$el.width(600);
        this.$el.height(400);
        _.each(this.$el.find("canvas"), function(c) {
            $(c).attr("width",
              (that.$el.find("div[aria-hidden=false]").width()) - 10);
            $(c).css("width", $(c).attr("width"));

            $(c).attr("height",
              (that.$el.find("div[aria-hidden=false]").height()) - 10);
            $(c).css("height", $(c).attr("height"));
        });
        this.$el.tabs("refresh");
    },
});

var SheetNameView = Backbone.View.extend({
    tagName: "li",
    initialize: function(){
        this.listenTo(this.model, "remove", this.remove);
    },
    showit: function(obj){
        console.log("sheet event:", obj);
    },
    render: function() {
        this.$el.append(_.template(
          '<a href="#<%= name %>_sheet"><%= name %></a>', this.model.toJSON()));
        return this;
    },
});
var SheetBodyView = Backbone.View.extend({
  tagName: "div",
  id: function() {return this.model.get("name") + "_sheet";},
  initialize: function() {
    this.listenTo(this.model, "remove", this.remove);
  },
  events:{
    "mousemove": "onMousemove",
    "mouseup"  : "onMouseup",
  },
  onMousemove: function(e) {
    this.model.trigger("updatePartialLink",
      relativePosFrom(e, this.$el, "update"));
  },
  onMouseup: function(e) {
    this.model.trigger("deletePartialLink", e);
  },
  render: function() {
    this.$el.append(_.template(
      '<canvas id="<%= name %>_canvas"></canvas>', this.model.toJSON()));
    return this;
  }
});

function relativePosFrom(fr, to, msg) {
  var anythingToArray = function(target) {
    if (_.has(target, "pageX")) {
      return [target.pageX, target.pageY];
    }
    else if (_.isArray(target)) {
      return target;
    }
    else if (_.has(target, "left")) {
      return [target.left, target.top];
    }
    else if (target.offset && _.isFunction(target.offset)) {
      return [target.offset().left, target.offset().top];
    }
  };

  fr = anythingToArray(fr);
  to = anythingToArray(to);

  // if (msg) {
  //   console.log(msg);
  //   console.log("from :", fr);
  //   console.log("to   :", to);
  // }
  return [fr[0] - to[0], fr[1] - to[1]];
}

var AddSheetButtonView = Backbone.View.extend({
  el: $("#addSheetButton"),
  initialize: function() {
    $("#addSheetDialog").dialog({
      autoOpen: false,
      modal: true,
      buttons: {
        "OK": function () {
          Sheets.add({name: $(this).find("input").val(),
                type: $(this).find("select").val()});
          $(this).dialog("close");
        },
        "Cancel": function() {
          $(this).dialog("close");
        }
      }
    });
    this.$el.button();
  },
  events: {
    "click": "addSheet"
  },
  addSheet: function() {
    $("#addSheetDialog").dialog("open");
  }
})


// Nodes

var NodesView = Backbone.View.extend({
    className: "nodes",
    initialize: function() {
        this.listenTo(this.collection, "add", this.addNodeView);
    },
    addNodeView: function(node) {
        var view = (new NodeView({model: node})).render();
        this.$el.append(view.el);
    },
});

var NodeView = Backbone.View.extend({
    className: "node",
    id: function() {return this.model.get("id");},
    events: {
        "dblclick": "paramOpen",
        "mousedown .terminalcircle": "addPartialLink",
        "mouseup   .terminalcircle": "connectPartialLink"
    },
    initialize: function(){
        this.listenTo(this.model, "remove", this.remove);
        this.listenTo(this.model, "change", this.render);
    },
    paramOpen: function() {
        $("#" + this.model.get("id") + delim + "param").dialog("open");
        console.log("#" + this.model.get("id") + delim + "param");
    },
    render: function() {
        var forrender = _.extend(this.model.toJSON(),
                     {delim: delim});
        var that = this;

        this.$el.html(_.template(_render("node"), forrender));
        this.$el.css("position", "absolute");
        this.$el.css("left", this.model.get("pos")[0] + 100 + "px");
        this.$el.css("top",  this.model.get("pos")[1] + 100 + "px");

        this.$el.draggable({
            cancel: ".terminalcircle",
            drag: function(){
                var newpos = relativePosFrom($(this),
                            $("#" + that.model.get("sheet") + "_canvas"));
                console.log($(this).offset(), newpos);
                that.model.set("pos", newpos);
            }
        });
        this.$el.children(".paramDialog").dialog({
            autoOpen: false,
            modal: true,
            open: function() {
                _.each(that.model.toJSON().parameters, function(p) {
                    $(this).find("input[name=" + p.name + "]").val(p.value);
                    $(this).find("selector[name=" + p.name + "_type]").val(p.type);
                });
            },
            buttons: {
                "OK": function() {
                    // update values
                    _.each($(this).find("input"), function(i) {
                        var obj = _.find(that.model.get("parameters"),
                                         function(p) {return p.name === $(i).attr("name");});
                        if (obj) obj.value = $(i).val();
                    });

                    // update types
                    _.each($(this).find("select"), function(i) {
                        var obj = _.find(that.model.get("parameters"),
                                         function(p) {return p.name + "_type" === $(i).attr("name");});
                        if (obj) obj.type = $(i).val();
                    });
                    $(this).dialog("close");
                },
                "Cancel": function() {$(this).dialog("close");}
            }
        });
        return this;
    },
    addPartialLink: function(e) {
        var activeSheet = $(".ui-tabs-panel[aria-hidden=false]");
        var term = $(e.target).parentsUntil(".terminals").parent();

        if(term.hasClass("interminals")){
            Sheets.trigger("addPartialLinkToActiveSheet",
                new Link({
                    fromNodeId: this.model.get("id"),
                    fromTermName: $(e.target).siblings(".name").text().trim(),
                    fromPos: relativePosFrom($(e.target), activeSheet),
                    toPos: relativePosFrom($(e.target), activeSheet),
                    isPartial: "from"
                })
            );
        } else {
            Sheets.trigger("addPartialLinkToActiveSheet",
                new Link({
                    toNodeId: this.model.get("id"),
                    toTermName: $(e.target).siblings(".name").text().trim(),
                    toPos: relativePosFrom($(e.target), activeSheet),
                    fromPos: relativePosFrom($(e.target), activeSheet),
                    isPartial: "to"
                })
            );
        }
    },
    connectPartialLink: function(e){
        var activeSheet = $(".ui-tabs-panel[aria-hidden=false]");
        Sheets.trigger("connectPartialLink", {
            toTermName: $(e.target).siblings(".name").text().trim(),
            toNodeId  : this.model.get("id"),
            toPos     : relativePosFrom($(e.target), activeSheet),
            fromTermName: $(e.target).siblings(".name").text().trim(),
            fromNodeId  : this.model.get("id"),
            fromPos     : relativePosFrom($(e.target), activeSheet),
        });
    }
});


// Links

var LinksView = Backbone.View.extend({
  initialize: function() {
    this.listenTo(this.collection, "add", this.addLink);
    this.listenTo(this.collection, "change", this.render);
    this.listenTo(this.collection, "remove", this.remove);
  },
  remove: function() {
    this.render();
  },
  render: function() {
    if (this.ctx !== undefined) {
      this.clear();
      this.collection.each(function(c) {
        c.set("ctx", this.ctx);
        c.set("can", this.can);
        (new LinkView({model: c})).render();
      }, this);
    } else{
        // console.log("ctx is undefined");
    }
    return this;
  },
  clear: function() {
    if (this.ctx) {
      this.ctx.clearRect(0, 0, this.can.width, this.can.height);
    }
  },
  updatePos: function(nodes){
    var that = this;
    nodes.each(function(n){
        that.collection.each(function(l){
            console.log("updating", n.get("id"), l.toJSON());
            if(n.get("id") === l.get("fromNodeId")){
                l.set("fromPos", n.get("fromPos"));
                    // relativePosFrom($("#" + l.get("fromTermId") + " .terminalcircle"),
                    //                 $("#" + l.get("sheet") + "_canvas")));
            } else
            if(n.get("id") === l.get("toNodeId")){
                l.set("toPos", n.get("toPos"));
                    // relativePosFrom($("#" + l.get("toTermId") + " .terminalcircle"),
                    //                 $("#" + l.get("sheet") + "_canvas")));
            }
        });
    });
    this.render();
  },
  addLink: function(link) {
    if (!this.ctx) {
        console.log("getting element of #" + link.get("sheet") + "_canvas");
        this.can = document.getElementById(link.get("sheet") + "_canvas");
        this.ctx = this.can.getContext('2d');
    }
    this.render();
  }
});

var LinkView = Backbone.View.extend({
    events: {
    },
    initialize: function() {
        this.model.get("ctx").strokeStyle = "black";

        if(this.model.get("fromPos") === undefined){
            this.model.set("fromPos", relativePosFrom(
                $("#" + this.model.get("fromTermId")),
                $("#" + this.model.get("sheet") + "_canvas")
                ));
        }
        if(this.model.get("toPos") === undefined){
            this.model.set("toPos", relativePosFrom(
                $("#" + this.model.get("toTermId")),
                $("#" + this.model.get("sheet") + "_canvas")
                ));
        }

        this.listenTo(App, "resized", this.render);
        this.listenTo(this.model, "remove", this.remove);
    },
    render:function () {
        console.log(this.model.get("fromPos"), this.model.get("toPos"));
        this.draw(this.model.get("fromPos"), this.model.get("toPos"));
        return this;
    },
    draw: function(fr, to) {
        fr = [Math.floor(fr[0]), Math.floor(fr[1])];
        to = [Math.floor(to[0]), Math.floor(to[1])];

        this.model.get("ctx").beginPath();
        // this.model.get("ctx").moveTo(fr[0], fr[1]);
        // this.model.get("ctx").lineTo(to[0], to[1]);
        this.intelligentLink(fr, to);
        this.model.get("ctx").stroke();
        this.model.get("ctx").closePath();
    },
    intelligentLink: function (fr, to){
        var mleft = (fr[0] + to[0]) / 2.0;
        var mtop = (fr[1] + to[1]) / 2.0;
        if (fr[0] < to[0]){
            var m1 = [mleft, fr[1]];
            var m2 = [mleft, (fr[1] + to[1]) / 2.0];
            var m3 = [mleft, to[1]];

            this.Bezier([fr, m1, m2]);
            this.Bezier([m2, m3, to]);
        }
        else{
            var margin = Math.min(Math.abs(fr[0] - to[0]), 50);
            var poses = [
                fr,
                [fr[0] + margin, fr[1]],
                [fr[0] + margin, (fr[1] + mtop) / 2.0],
                [fr[0] + margin, mtop],
                [mleft,          mtop],
                [to[0] - margin, mtop],
                [to[0] - margin, (to[1] + mtop) / 2.0],
                [to[0] - margin, to[1]],
                to
            ];

            this.Bezier([poses[0], poses[1], poses[2]]);
            this.Bezier([poses[2], poses[3], poses[4]]);
            this.Bezier([poses[4], poses[5], poses[6]]);
            this.Bezier([poses[6], poses[7], poses[8]]);
        }
    },
    Bezier: function(poses){
        /*http://ruiueyama.tumblr.com/post/11197882224*/
        function interp(a, b, f){return a + f * (b - a);}
        var split = 15;

        this.model.get("ctx").moveTo(poses[0][0], poses[0][1]);
        for(var i=0; i<split; i++){
            var p1 = [interp(poses[0][0], poses[1][0], 1.0 * i / split),
                      interp(poses[0][1], poses[1][1], 1.0 * i / split)];
            var p2 = [interp(poses[1][0], poses[2][0], 1.0 * i / split),
                      interp(poses[1][1], poses[2][1], 1.0 * i / split)];

            this.model.get("ctx").lineTo(
                interp(p1[0], p2[0], 1.0 * i / split),
                interp(p1[1], p2[1], 1.0 * i / split));
        }
        this.model.get("ctx").lineTo(poses[2][0], poses[2][1]);
    }
})

// TerminalRoles

// error
var ErrorMessage = Backbone.View.extend({
    el: $("#errorMessage"),
    initialize: function(){
        this.listenTo(Sheets, "invalid", function(msg){this.$el.text(msg)});
    }
});


// menubar
var Sidebar = Backbone.View.extend({
    el: $("#sidebar"),
    initialize: function(){
        this.listenTo(Nodedefs, "change", this.render);
        this.listenTo(Nodedefs, "add", this.render);
        this.render();
    },
    render: _.throttle(function(){
        if(!this.$el.hasClass("ui-accordion")){
            this.$el.accordion();
        }else {
            this.$el.accordion("refresh");
        }
        this.$el.children(".ui-accordion-content").css("padding", "0px");
        this.$el.width()
        var maxwidth = _.chain(this.$el.children(".ui-accordion-content"))
                        .map(function(c){return parseInt($(c).css("width"));})
                        .max().value();
        _.each(this.$el.children(".ui-accordion-content"),
            function(c){$(c).css("width", maxwidth + "px");});
    }, 500)
});


// file
var FileView = Backbone.View.extend({
    tagName: "li",
    events: {
      "click button[name=load]"  : "load",
      "click button[name=delete]": "delete",
    },
    initialize: function(){
        this.listenTo(this.model, "destroy", this.remove);
    },
    render: function(){
        this.$el.append(_.template(
            '<span name="<%= name %>"><%= name %></span>',
            this.model.toJSON()));
        this.$el.append('<button name="load">load</button>');
        this.$el.append('<button name="delete">delete</button></li>');
        return this;
    },
    load: function(){
        Sheets.trigger("replace", this.model.get("object"));
    },
    delete: function(){
        console.log("destroy:", this.model.isNew());
        this.model.destroy();
    }
});
var FileListView = Backbone.View.extend({
    el: $("#filelist"),
    initialize: function(){
        this.listenTo(this.collection, "add", this.addFile);
        this.$el.siblings("#noFile").css("display", "none");
        this.collection.fetch({add: true});
    },
    addFile: function(f){
        var view = (new FileView({model: f})).render();
        console.log(f.toJSON());
        this.$el.append(view.el);
    },
});

var FileUploadView = Backbone.View.extend({
    el: $("#fileUpload"),
    initialize: function(){
        var that = this;
        this.$el.children("[name=path]").browse({
            input: this.$el.children("[name=name]"),
            button: this.$el.children("button")
        });

        this.$el.children("[name=path]").on("change", function(a){
            _.each(this.files, function(f){that.upload(f);});
        });
    },
    upload: function(fileobj){
        var reader = new FileReader();
        var that = this;
        reader.onload = function(e){
            var f = new File({
              name:    fileobj.name,
              content: e.target.result,
              type: "network"});
            f.save();
            f.fetch();
            that.collection.add(f);
        };
        reader.readAsText(fileobj);
    }
});

var RolesView = Backbone.View.extend({
    el: $("#sheets"),
    initialize: function(){
        this.listenTo(this.collection, "add", this.add);
        $("#terminalRoleDialog").dialog({
            autoOpen: false,
            modal: true,
            buttons: {
                "OK": function () {
                    Sheets.trigger("addTerminalRoleToSheets",
                                $(this).attr("type"),
                                $(this).attr("termId"),
                                $(this).attr("sheetId"),
                                $(this).find("input").val());
                    $(this).dialog("close");
                },
                "Cancel": function () {
                    $(this).dialog("close");
                }
            }
        });
        this.$el.contextMenu({
            selector: ".node .terminalcircle",
            build: function($trigger, e){
                var obj = {"items": {}};
                $("#terminalRoleDialog").attr("termId",
                    $trigger.parent().attr("id"));
                $("#terminalRoleDialog").attr("sheetId",
                    $trigger.parentsUntil("#sheets", ".ui-tabs-panel").attr("id"));
                if($trigger.parent().parent().hasClass("interminals")){
                    obj["items"]["Set as Input"] = {
                        name: "Set as Input", icon: "edit",
                        callback: function(key, opts){
                            $("#terminalRoleDialog").attr("type", "input");
                            $("#terminalRoleDialog").dialog("open");
                        }
                    }
                }
                else{
                    obj["items"]["Set as Output"] = {
                        name: "Set as Output", icon: "edit",
                        callback: function(key, opts){
                            $("#terminalRoleDialog").attr("type", "output");
                            $("#terminalRoleDialog").dialog("open");
                        }
                    }
                    obj["items"]["Set as Condition"] = {
                        name: "Set as Condition", icon: "edit",
                        callback: function(key, opts){
                        Sheets.trigger("addTerminalRoleToSheets",
                                    "condition",
                                    $("#terminalRoleDialog").attr("termId"),
                                    $("#terminalRoleDialog").attr("sheetId"), "CONDITION");
                        }
                    }
                }
                return obj;
            }
        });
    },
    add: function(role){
        var view = (new RoleView({model: role})).render();
    }
});

var RoleView = Backbone.View.extend({
    initialize: function(){
        this.listenTo(this.model, "remove", this.clearText);
    },
    clearText: function(){
        var obj = $("#" + this.model.get("nodeId") + delim + this.model.get("termName"));
        obj.children(".terminalcircle").children(".terminalRoleName")
           .text("");
    },
    render: function(){
        var obj = $("#" + this.model.get("nodeId") + delim + this.model.get("termName"));
        obj.children(".terminalcircle").children(".terminalRoleName")
           .text(this.model.get("name"));
    }
});